
// DepTree.cpp
// Copyright 2022 Matthew Rickard
// This file is part of dep

// dep is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.

// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <https://www.gnu.org/licenses/>.

#include "DepTree.h"
#include "depparse.h"
#include "logMsg.h"
// Extra headers (user code)
#include "quote.h"
#include <string.h>
#include "AbstractProcess.h"
#include "MacroNames.h"
#include "stringify.h"
#include <unistd.h>
#include "matches.h"
#include "clockwork.h"

DepTree *depTree; // (wizard code)

DepTree::DepTree():
    ParseTree(dep_create_buffer, dep_scan_string, dep_switch_to_buffer) { // (customizable code - constructor)
    // constructor (user code)
  activeCode = 1;
  rule = 0;
  log_lexemes = 0;
  inAMacro = 0;
}

int DepTree::callParser(YY_BUFFER_STATE ybs) { // (wizard code)
  deplloc.first_line = 1;
  deplloc.first_column = 1;
  last_line = 1;
  last_column = 1;
  depTree = this;
  lexemes = 0;
  error = 0;
  depdebug = getenvint("DEBUG_PARSER");
  symbolToNameFnPtr = depparseGetSymbolName;
  // callParser (user code)
  int result = depparse(); // (wizard code)
  if (!result)
    result = error;
  // Cleanup code (user code)
  return result; // (wizard code)
}

int DepTree::processToken(int token) {
  ParseTree::updatePosition(&deplloc, deptext);
  static int log_lexemes = getenvint("LOG_LEXEMES");
  if ((log_lexemes || this->log_lexemes)) {
    std::string symbolName = depparseGetSymbolName(depparseGetSymbolNumber(token));
    if (ruleset->outputFile->getOStream())
      *ruleset->outputFile->getOStream() << symbolName << " (" << filename << ' '
        << deplloc.first_line << '.' << deplloc.first_column << ":"
        << " \"" << quote(deptext) << "\")\n";
  }
  lexemes++;
  deplval = new LeafNode(depparseGetSymbolNumber(token), deptext, filename.c_str(), deplloc); // (customizable code - new LeafNode)
  return token;
}

int DepTree::popBuffer() {
  // (user code)
  if (!sourceFileVector.size()) {
    return 1;
  }
  if (deplexGetCurrentFile()) {
    fclose(deplexGetCurrentFile());
  }
  dep_delete_buffer(deplexGetCurrentBuffer());
  dep_switch_to_buffer(sourceFileVector.back().bufferState);
  this->filename = sourceFileVector.back().filename;
  if (ruleset->otherVerbosity >= 1)
    if (!this->inAMacro)
      *ruleset->outputFile->getOStream() << "Returning to " << filename << std::endl;
  depReaderCurrentFolder = sourceFileVector.back().folder;
  deplloc = sourceFileVector.back().location;
  last_line = sourceFileVector.back().location.first_line;
  last_column = sourceFileVector.back().location.first_column;
  inAMacro = sourceFileVector.back().inAMacro;
  sourceFileVector.pop_back();
  return 0;
}

void DepTree::callMacro(const TreeNode *node) {
  std::string str(node->str());
  if (variableMap.count(str)) {
    pushBuffer(filename + " (macro " + str + ")", 1);
    dep_scan_string(variableMap[str].c_str());
  }
  else {
    const char *value = getenv(str.c_str());
    if (value) {
      pushBuffer(filename + " env " + str, 1);
      //logMsg << "Calling an env" << std::endl;
      dep_scan_string(value);
    }
  }
}

void DepTree::substituteCommand(const TreeNode *node) {
  std::string command(node->str());
  Process process(command);
  process.pipeOutput();
  process.run();

  pushBuffer(filename + " cmdsubst " + command, 1);
  std::string output;
  ReadWholeFile(process.outputPipe, output);
  dep_scan_string(output.c_str());
}

void DepTree::includeFile(const TreeNode *node, int nocd) {
  if (!activeCode)
    return;
  // Keep good track of the filenames!
  //logMsg << logValue(depReaderCurrentFolder) << std::endl;
  //logMsg << logValue(node->str()) << std::endl;
  std::string filename = joinFolderAndFile(depReaderCurrentFolder, node->str());
  if (ruleset->otherVerbosity >= 1)
    *ruleset->outputFile->getOStream() << "Including " << filename /*<< ' ' << logValue(nocd)*/ << std::endl;
  FILE *file = fopen(filename.c_str(), "r");
  if (file) {
    pushBuffer(this->filename + ':' + stringify(last_line) + " include of " + filename, 0);
    //logMsg << logValue(file) << std::endl;
    if (!nocd) {
      //logMsg << logValue(depReader->currentFolder) << std::endl;
      //depReader->currentFolder = joinFolderAndFile(depReader->currentFolder,
        //DirName(filename.c_str()));
      depReaderCurrentFolder = DirName(filename.c_str());
      //logMsg << logValue(depReader->currentFolder) << std::endl;
    }
    dep_switch_to_buffer(dep_create_buffer(file, YY_BUF_SIZE));
  }
  else {
    int err = errno;
    logMsg << "dep: cannot read " << filename;
    if (err == ENOENT)
      std::cerr << " File does not exist" << std::endl;
    else {
      std::cerr << " error is " << ErrnoToMacroName(err) << ' ' << strerror(err) << std::endl;
    }
  }
}

// Some day: Move as much of this as possible to ParseTree

void DepTree::pushBuffer(const std::string &newFilename, int inAMacro) {
  SourceFile sourceFile;
  sourceFile.bufferState = deplexGetCurrentBuffer();
  sourceFile.filename = filename;
  sourceFile.location.first_line = last_line;
  sourceFile.location.first_column = last_column;
  sourceFile.inAMacro = this->inAMacro;
  sourceFile.folder = depReaderCurrentFolder;
  sourceFileVector.push_back(sourceFile);
  filename = newFilename;
  this->inAMacro = inAMacro;
  //logMsg << logValue(deplloc.first_line) << std::endl;
  if (!inAMacro) {
    deplloc.first_line = 1;
    deplloc.first_column = 1;
    last_line = 1;
    last_column = 1;
  }
  //logMsg << logValue(newFilename) << std::endl;
}

void DepTree::setVariable(const TreeNode *variable, const TreeNode *value, int env) {
  if (activeCode) {
    if (ruleset->otherVerbosity >= 6)
      *ruleset->outputFile->getOStream() << logValue(activeCode) << ' '
        << variable->str() << '=' << value->str() << std::endl;
    //logMsg << *variable->getPosition() << ": " << variable->str() << '=' << value->str() << std::endl;
    variableMap[variable->str()] = value->str();
    if (variable->str() == "line_") {
      last_line = atoi(value->str().c_str());
      //logMsg << "Changing line to " << last_line << std::endl;
    }
    if (env)
      setenv(variable->str().c_str(), value->str().c_str(), 1);
  }
}

int DepTree::evaluateExpr(const TreeNode *node) {
  std::string grammarRule = node->getRule();
  int result;
  if (grammarRule == "expr: expr BARBAR ws.opt expr")
    result = evaluateExpr(node->subNode(0)) || evaluateExpr(node->subNode(3));
  else if (grammarRule == "expr: expr '=' ws.opt expr") {
    //logMsg << '"' << node->subNode(0)->str() << '"' << std::endl;
    //logMsg << '"' << node->subNode(3)->str() << '"' << std::endl;
    result = stripWhitespace(node->subNode(0)->str()) == stripWhitespace(node->subNode(3)->str());
  }
  else if (grammarRule == "expr: '=' ws.opt expr")
    result = node->subNode(2)->str() == "";
  else if (grammarRule == "expr: '!' ws.opt expr")
    result = !evaluateExpr(node->subNode(2));
  else if (grammarRule == "expr: '!' ws.opt")
    result = 1;
  else if (grammarRule == "expr: argument ws.opt")
    result = evaluateExpr(node->subNode(0));
  else if (grammarRule == "argument: argumentPiece")
    result = evaluateExpr(node->subNode(0));
  else if (grammarRule == "argumentPiece: EVERYTHINGELSE")
    result = evaluateExpr(node->subNode(0));
  else if (grammarRule == "EVERYTHINGELSE") {
    result = atoi(node->str().c_str());
    //logMsg << logValue(node->str()) << std::endl;
    //logMsg << logValue(result) << std::endl;
  }
  else {
    logMsg << "Don't understand " << logValue(grammarRule) << " at " << *node->getPosition()
      << std::endl;
    lineabort();
  }
  //logMsg << logValue(grammarRule) << std::endl;
  //logMsg << logValue(node->str()) << std::endl;
  //logMsg << logValue(result) << std::endl;

  if (ruleset->otherVerbosity >= 1) {
    //*ruleset->outputFile->getOStream() << node->str(HIDDEN_MACRO_CALL) << ' ' << logValue(result) << std::endl;
  }

  return result;
}

void DepTree::expandForLoop(const Position *pos, const std::string &identifier,
    const TreeNode *node, const TreeNode *ws, const TreeNode *cmd, std::ostringstream &os) {
  std::string grammarRule = node->getRule();
  if (grammarRule == "argumentList: argumentList argumentPiece ws.opt") {
    expandForLoop(pos, identifier, node->subNode(0), ws, cmd, os);
    expandForLoop(pos, identifier, node->subNode(1), ws, cmd, os);
  }
  else if (grammarRule == "argumentList:");
  else if (grammarRule == "argumentPiece: IDENTIFIER"
        || grammarRule == "argumentPiece: STRING") {
    std::string backup;
    int count = variableMap.count(identifier);
    if (count)
      backup = variableMap[identifier];

    // This is a whitespace-separated list expander
    std::string value1 = node->str(HIDDEN_MACRO_CALL);
    std::string value;
    skipAndDequote(value1.c_str(), &value);
    //logMsg << logValue(value1) << std::endl;
    //logMsg << logValue(value) << std::endl;
    std::string dequoted;
    const char *s = value.c_str();
    while (s && *s) {
      s = skipAndDequote(s, &dequoted);
      //logMsg << logValue(dequoted) << std::endl;
      os << identifier << '=' << dequoted << '\n';
      os << "line_=" << pos->line << '\n';
      os << ws->str(HIDDEN_MACRO_CALL);
      os << cmd->str(HIDDEN_MACRO_CALL);
      s = skipWhitespace(s);
    }

    if (count)
      os << identifier << '=' << backup << '\n';
    else
      os << identifier << "=\n";
  }
  else {
    logMsg << "Don't understand " << logValue(grammarRule) << std::endl;
    lineabort();
  }
}

const char *fragmentTypeToMacroName(int t) {
  if (t == FRAGMENTTYPE_OPTION) return "OPTION";
  if (t == FRAGMENTTYPE_PARAM) return "PARAM";
  if (t == FRAGMENTTYPE_COMMAND) return "PARAM";
  return "(Unknown fragment type)";
}

struct DepCmdLine: CmdLine {
  DepCmdLine(const char *currentFolder, Rule *rule):
    CmdLine(0), flag(0), paramType(0), /*findIncludesFlag(0),*/ depline(0),
      rule(rule), currentFolder(currentFolder), dclFilename("b1ab1a") {
  }
  int flag;
  int paramType;
  void processFragment(const Fragment *fragment) {
    //logMsg << logValue(dclFilename) << std::endl;
    //logMsg << fragment->optionValue << ' ' << fragmentTypeToMacroName(fragment->fragmentType) << ' '
      //<< !!fragment->option << ' ' << (fragment->argValue? fragment->argValue: "(null)")
      //<< ' ' << logValue(depline) << std::endl;
    if (fragment->fragmentType == FRAGMENTTYPE_OPTION
     || fragment->fragmentType == FRAGMENTTYPE_PARAM && depline == 0) {
      if (fragment->fragmentType == FRAGMENTTYPE_PARAM) {
        //logMsg << logValue(fragment->argValue) << ' ' << logValue(paramType) << std::endl;
        if (!strcmp(fragment->argValue, "%in"))
          paramType = 1;
        else if (!strcmp(fragment->argValue, "%out"))
          paramType = 2;
        else if (!strcmp(fragment->argValue, ">"))
          paramType = 2;
        else if (!strcmp(fragment->argValue, "%cmpout"))
          paramType = 3;
        else if (!strcmp(fragment->argValue, "%pseudo"))
          paramType = 4;
        else if (!strcmp(fragment->argValue, "%exist"))
          paramType = 5;
        else if (!strcmp(fragment->argValue, "%outz"))
          paramType = 6;
        else if (paramType == 1)
          rule->addSource(fragment->argValue, 0);
        else if (paramType == 2)
          rule->addTarget(fragment->argValue, 0, dclFilename, dclLine);
        else if (paramType == 3)
          rule->addTarget(fragment->argValue, FILEFLAG_COMPARE, dclFilename, dclLine);
        else if (paramType == 4)
          rule->addTarget(fragment->argValue, FILEFLAG_PSEUDOFILE, dclFilename, dclLine);
        else if (paramType == 5)
          rule->addSource(fragment->argValue, FILEFLAG_EXIST);
        else if (paramType == 6)
          rule->addTarget(fragment->argValue, FILEFLAG_ZEROLENGTH, dclFilename, dclLine);
      }
      if (paramType == 0 && fragment->option) {
        //logMsg << logValue(*fragment) << std::endl;
        if (fragment->option->primary->argName == "%includedir") {
          //logMsg << "Found a %includedir" << std::endl;
          includeDirList.push_back(fragment->argValue);
        }
        else if (fragment->option->primary->argName == "%libdir") {
          // Speedup: eliminate values that are absolute paths
          if (fragment->argValue && *fragment->argValue != '/')
            libDirList.push_back(fragment->argValue);
        }
        else if (fragment->option->primary->argName == "%library") {
          libFileList.push_back(fragment->argValue);
        }
        else if (fragment->option->primary->argName == "%in") {
          if (depTree->ruleset->otherVerbosity >= 5) {
            *depTree->ruleset->outputFile->getOStream() << "dep: adding an input "
              << logValue(fragment->argValue) << std::endl;
          }
          rule->addSource(fragment->argValue, 0);
        }
        else if (fragment->option->primary->argName == "%exist") {
          if (depTree->ruleset->otherVerbosity >= 5) {
            *depTree->ruleset->outputFile->getOStream() << "dep: adding an existence input "
              << logValue(fragment->argValue) << std::endl;
          }
          rule->addSource(fragment->argValue, FILEFLAG_EXIST);
        }
        else if (fragment->option->primary->argName == "%out") {
          rule->addTarget(fragment->argValue, 0, dclFilename, dclLine);
        }
        else if (fragment->option->primary->argName == "%cmpout") {
          rule->addTarget(fragment->argValue, FILEFLAG_COMPARE, dclFilename, dclLine);
        }
        else if (fragment->option->primary->argName == "%outz") {
          rule->addTarget(fragment->argValue, FILEFLAG_ZEROLENGTH, dclFilename, dclLine);
        }
        else if (fragment->option->primary->argName == "%pseudo") {
          rule->addTarget(fragment->argValue, FILEFLAG_PSEUDOFILE, dclFilename, dclLine);
        }
        //else if (fragment->option->primary->argName == "%findincludes") {
          //*depTree->ruleset->outputFile->getOStream() << "Setting findIncludesFlag" << std::endl;
          //logMsg << "Setting findIncludesFlag" << std::endl;
          //findIncludesFlag = 1;
        //}
        //else
          //logMsg << logValue(fragment->option->primary->argName) << std::endl;
      }
    }
    else if (fragment->fragmentType == FRAGMENTTYPE_PARAM) {
      if (depline == 0) {
        lineabort();
      }
      else if (depline) {
        //logMsg << "checkpoint" << std::endl;
        if (!strcmp(fragment->argValue, "%in"))
          depline = 1;
        else if (!strcmp(fragment->argValue, "%out"))
          depline = 2;
        else if (!strcmp(fragment->argValue, "%cmpout"))
          depline = 3;
        else if (!strcmp(fragment->argValue, "%pseudo"))
          depline = 4;
        else if (!strcmp(fragment->argValue, "%exist"))
          depline = 5;
        else if (!strcmp(fragment->argValue, "%outz"))
          depline = 6;
        else if (depline == 1)
          rule->addSource(fragment->argValue, 0);
        else if (depline == 2)
          rule->addTarget(fragment->argValue, 0, dclFilename, dclLine);
        else if (depline == 3)
          rule->addTarget(fragment->argValue, FILEFLAG_COMPARE, dclFilename, dclLine);
        else if (depline == 4)
          rule->addTarget(fragment->argValue, FILEFLAG_PSEUDOFILE, dclFilename, dclLine);
        else if (depline == 5)
          rule->addSource(fragment->argValue, FILEFLAG_EXIST);
        else if (depline == 6)
          rule->addTarget(fragment->argValue, FILEFLAG_ZEROLENGTH, dclFilename, dclLine);
      }
    }
    else if (fragment->fragmentType == FRAGMENTTYPE_COMMAND) {
      if (!strcmp(fragment->argValue, "%in")) {
        depline = 1;
        flag = 0;
      }
      else if (!strcmp(fragment->argValue, "%out")) {
        depline = 2;
        flag = 0;
      }
      else if (!strcmp(fragment->argValue, "%cmpout")) {
        depline = 3;
        flag = FILEFLAG_COMPARE;
      }
      else if (!strcmp(fragment->argValue, "%pseudo")) {
        depline = 4;
        flag = FILEFLAG_PSEUDOFILE;
      }
      else if (!strcmp(fragment->argValue, "%exist")) {
        depline = 5;
        flag = FILEFLAG_EXIST;
      }
      else if (!strcmp(fragment->argValue, "%outz")) {
        depline = 6;
        flag = FILEFLAG_ZEROLENGTH;
      }
      else if (depTree->ruleset->syntaxMap.count(fragment->argValue)) {
        cmdspec = depTree->ruleset->syntaxMap[fragment->argValue];
      }
      else if (!strcmp(fragment->argValue, "export"));
      else {
        //logMsg << logValue(rule->ruleset->otherVerbosity) << std::endl;
        if (rule->ruleset->otherVerbosity >= 0)
          errorLine = std::string("dep: ") + dclFilename + " " + stringify(dclLine)
            + ": unknown syntax for " + fragment->argValue;
          cmdspec = 0;
      }
    }
  }
  std::vector<std::string> includeDirList;
  std::vector<std::string> libDirList;
  std::vector<std::string> libFileList;
  //int findIncludesFlag;
  int depline;
  Rule *rule;
  const char *currentFolder;
  std::string errorLine;
  const char *dclFilename;
  int dclLine;
};

// This very new version reads summarized includes from a DepFile

void DepTree::fastAddFile(const std::string &filename, const TreeNode *buildStatus,
    const TreeNode *timestamp, const TreeNode *argumentList) {
  //logMsg << "Hello from fastAddFile" << std::endl;
  BuildFile *buildFile = ruleset->getBuildFile(filename.c_str());
  std::string rule = timestamp->getRule();
  if (rule == "timestamp: EVERYTHINGELSE ws.opt EVERYTHINGELSE ws.opt EVERYTHINGELSE ws.opt") {
    std::string ts = timestamp->subNode(0)->str() + " " + timestamp->subNode(2)->str();
    int year, mon, day, hour, min;
    Decimal sec;
    std::string tzname, tz;
    scanDatetime("y-m-d h:m:sz", ts.c_str(), &year, &mon, &day,
      &hour, &min, &sec, &tzname, &tz);
    changeOffset(&year, &mon, &day, &hour, &min, &sec, &tzname, &tz, "UTC", "+0");
    encodeDatetime(year, mon, day, hour, min, (int)sec,
      UNIX_EPOCH, 60, &buildFile->localModTime.tv_sec);
    buildFile->localModTime.tv_nsec = (sec - (int)sec) * 1000000000;
    buildFile->fileflag |= FILEFLAG_LOCAL_MODTIME_VALID;
    buildFile->modTimeResult = 0;
    buildFile->fileSize = atol(timestamp->subNode(4)->str().c_str());
  }
  else if (rule == "timestamp: IDENTIFIER ws.opt") {
    buildFile->localModTime.tv_sec = 0;
    buildFile->localModTime.tv_nsec = 0;
    std::string ts = timestamp->subNode(0)->str();
    if (ts == "unknown") {
      buildFile->fileflag &= ~FILEFLAG_LOCAL_MODTIME_VALID;
      buildFile->modTimeResult = 0;
    }
    else if (ts == "ENOENT") {
      buildFile->fileflag |= FILEFLAG_LOCAL_MODTIME_VALID;
      buildFile->modTimeResult = ENOENT;
    }
    else if (ts == "EINVAL") {
      buildFile->fileflag |= FILEFLAG_LOCAL_MODTIME_VALID;
      buildFile->modTimeResult = EINVAL;
    }
    else {
      logMsg << "Can't do " << logValue(rule) << ' ' << logValue(ts) << std::endl;
      lineabort();
    }
  }
  else {
    logMsg << "Can't do " << logValue(rule) << ' ' << std::endl;
    logMsg << logValue(timestamp->str()) << std::endl;
    lineabort();
  }
  fastAddIncludes(buildFile, argumentList);
}

// fastAddIncludes() - process include directives parsed from a .DepFileCache file

void DepTree::fastAddIncludes(BuildFile *buildFile, const TreeNode *node) {
  //logMsg << "Hello from fastAddIncludes" << std::endl;
  std::string grammarRule = node->getRule();
  if (grammarRule == "argumentList: argumentList argumentPiece ws.opt") {
    Cmdspec cmdspec("", 0, *node->getPosition());
    CmdLineArgvMaker cmdline(CMDPACK_UNPACK, &cmdspec);
    cmdline.processCmdline(FRAGMENTTYPE_PARAM, node->str().c_str(), ' ');
    for (std::vector<char *>::iterator i = cmdline.v.begin(); i != cmdline.v.end(); i++) {
      //logMsg << "Adding an include" << std::endl;
      buildFile->treeAddInclude(*i);
    }
  }
  else if (grammarRule == "includes:");
  else if (grammarRule == "includes: KEYWORD_INCLUDES ws.opt argumentList") {
    fastAddIncludes(buildFile, node->subNode(2));
  }
  else {
    logMsg << "Don't understand " << logValue(grammarRule) << std::endl;
    lineabort();
  }
}

void DepTree::depTreeAddRule(const Position *pos) {
  //logMsg << "Adding rule " << std::endl;
  rule = ruleset->rulesetAddRule(pos);
  //lineabort();
  rule->setFolder(currentFolder);
}

void DepTree::depTreeAddCommand(std::string s, const Position *pos) {
  //logMsg << "Adding command " << logValue(s) << std::endl;
  size_t posn = s.find("#");
  if (posn != std::string::npos) {
    //logMsg << "Changing " << s << std::endl;
    s.resize(posn);
    //logMsg << "      to " << s << std::endl;
  }
  if (!rule)
    lineabort();
  for (;;) {
    size_t pos = s.find("\\\n");
    if (pos == std::string::npos)
      break;
    // Ways to handle backslash-newlines:
    // 1) erase them at the lexer
    // 2) erase them here (this is the current way)
    // 3) handle them specially in the argument scanner
    s.erase(pos, 2);
  }
  if (!s.size())
    return;

  //logMsg << logValue(lookForDependencies) << std::endl;
  if (lookForDependencies) {
    //*ruleset->outputFile->getOStream() << "Constructing a DepCmdLine " logValue(s) << std::endl;
    DepCmdLine cmdline(depReaderCurrentFolder.c_str(), rule);

    cmdline.dclFilename = pos->filename;
    cmdline.dclLine = pos->line;

    //logMsg << "Processing line " << logValue(s) << std::endl;
    cmdline.processCmdline(FRAGMENTTYPE_COMMAND, s.c_str(), DIVIDER_WHITESPACE);
    if (cmdline.errorLine.size() && !cmdline.paramType) {
      //logMsg << logValue(ruleset->os) << std::endl;
      if (ruleset->outputFile->getOStream())
        *ruleset->outputFile->getOStream() << cmdline.errorLine << std::endl;
    }
    rule->addCommand(s);

    //if (ruleset->otherVerbosity >= 3) // 3 OK 4 too high
      //*ruleset->outputFile->getOStream() << logValue(cmdline.findIncludesFlag) << std::endl;
    if ((s[0] != '%') /*|| cmdline.findIncludesFlag*/) {
      cmdline.includeDirList.push_back(""); // Put the empty include at the end of the list
      for (auto sv: rule->sourceVector) {
        for (std::vector<IncludeRule>::iterator ir = ruleset->includeRuleVector.begin();
            ir != ruleset->includeRuleVector.end(); ir++) {
          //logMsg << logValue(ir->syntax) << std::endl;
          //logMsg << logValue(ir->fileType.size()) << std::endl;
          for (std::vector<Pattern>::iterator ft = ir->filePatternVector.begin();
              ft != ir->filePatternVector.end(); ft++) {
            Pattern &pattern(*ft);
            int match = pattern.matches(sv->filename);
            //logMsg << logValue(sv->filename) << ' ' << logValue(sv->filename)
              //<< ' ' << logValue(match) << std::endl;
            if (ruleset->otherVerbosity >= 4)
              *ruleset->outputFile->getOStream() << logValue(sv->filename)
                << ' ' << logValue(match) << std::endl;
            if (match) {
              if (ruleset->otherVerbosity >= 3) // 3 OK 4 too high
                *ruleset->outputFile->getOStream() << "Looking in " << sv->filename << std::endl;
              std::fstream is;
              if (ruleset->otherVerbosity >= 5)
                *ruleset->outputFile->getOStream() << BaseName(__FILE__) << " Calling findFile on "
                  << sv->filename << std::endl;
              std::string foundFile(ruleset->findFile(sv->filename,
                &cmdline.includeDirList, is, "", rule->folder.c_str()));
              if (ruleset->otherVerbosity >= 5)
                *ruleset->outputFile->getOStream() << BaseName(__FILE__) << " Calling addIncludes on "
                  << logValue(foundFile) << std::endl;
              combinedAddIncludes(ruleset->optionTreeIncludes? sv: 0,
                ruleset->optionRuleIncludes? rule: 0, foundFile,
                &cmdline.includeDirList, is, rule->folder.c_str());
            }
          }
        }
      }
    }

    for (auto j: cmdline.libFileList) {
      int foundIt = 0;
      for (auto i: cmdline.libDirList) {
        std::string fullName = joinFolderAndFile(i, "lib" + j + ".so");
        if (ruleset->fileMap.count(fullName)) {
          rule->addSource(fullName, 0);
          foundIt = 1;
        }
        else {
          fullName = joinFolderAndFile(i, "lib" + j + ".dll");
          if (ruleset->fileMap.count(fullName)) {
            rule->addSource(fullName, 0);
            foundIt = 1;
          }
          else {
            fullName = joinFolderAndFile(i, "lib" + j + ".a");
            if (ruleset->fileMap.count(fullName)) {
              rule->addSource(fullName, 0);
              foundIt = 1;
            }
          }
        }
      }
      if (!foundIt) {
        if (ruleset->otherVerbosity >= 4) {
          *ruleset->outputFile->getOStream() << filename << ':' << deplloc.first_line
            << " Couldn't find build rule for " << j << std::endl;
          for (auto i: cmdline.libDirList) {
            std::string fullName = joinFolderAndFile(i, "lib" + j + ".so");
            if (!ruleset->fileMap.count(fullName)) {
              *ruleset->outputFile->getOStream() << "It's not " << fullName << std::endl;
            }
            fullName = joinFolderAndFile(i, "lib" + j + ".dll");
            if (!ruleset->fileMap.count(fullName)) {
              *ruleset->outputFile->getOStream() << "It's not " << fullName << std::endl;
            }
            fullName = joinFolderAndFile(i, "lib" + j + ".a");
            if (!ruleset->fileMap.count(fullName)) {
              *ruleset->outputFile->getOStream() << "It's not " << fullName << std::endl;
            }
          }
        }
      }
    }
  }
  else {
    //logMsg << "lookForDependencies is 0 so not looking" << std::endl;
    rule->addCommand(s);
  }
}

int DepTree::readFile(String filename) {
  if (ruleset->maxNonIncludes == -13)
    ruleset->maxNonIncludes = 1000;

  depTree = this;
  FILE *f = fopen(filename, "r");
  if (!f) {
    int err = errno;
    if (ruleset->otherVerbosity >= 1)
      *ruleset->outputFile->getOStream() << "Couldn't open " << filename.s << " for reading "
      << logValue(err) << ' ' << ErrnoToMacroName(err)
      << std::endl;
    return -1;
  }
  if (ruleset->otherVerbosity >= 1)
    *ruleset->outputFile->getOStream() << "Opened " << filename.s << " for reading" << std::endl;
  int result = parseOpenFile(filename, f);
  fclose(f);
  if (result) {
    logMsg << logValue(result) << ". Exiting now" << std::endl;
    return -1;
  }
  return 0;
}

int DepTree::runImmediate(String s) {
  if (!strncmp(s.s, "imm echo ", 9)) {
    //std::cout << s.s + 9 << '\n'; // Skip over the words "imm echo"
    *ruleset->outputFile->getIOStream() << s.s + 9 << '\n';
    return 1;
  }
  return 0;
}

void removeComment(std::string &s) {
  size_t hash = s.find('#');
  if (hash != std::string::npos) {
    //logMsg << logValue(s) << std::endl;
    s.resize(hash);
    s = stripWhitespace(s);
    //logMsg << logValue(s) << std::endl;
  }
}

